/* ====================================================================== */
/*                              vendor_info                               */
/* ====================================================================== */

/* 
 * vendor_info_array
 *
 * This array contains the list of Subsystem IDs on which the driver
 * should load.
 *
 * The format of each entry of the array is as follows:
 * { SUBSYSTEM_VENDOR, SUBSYSTEM_DEV, BRANDING_STRING }
 *
 * If there is a CATCHALL in the SUBSYSTEM_DEV field, the driver will
 * load on all subsystem device IDs of that vendor.
 * If there is a CATCHALL in the SUBSYSTEM_VENDOR field, the driver will
 * load on all cards, irrespective of the vendor.
 *
 * The last entry of the array must be:
 * { 0, 0, "" }
 */

#ifndef CATCHALL
#define CATCHALL 0xffff
#endif

#ifndef _VENDOR_INFO_T
#define _VENDOR_INFO_T
typedef struct _e1000_vendor_info_t
{
	uint16_t sub_ven;
	uint16_t sub_dev;
	char *idstr;					  /* String to be printed for these IDs */
}
e1000_vendor_info_t;
#endif /* _VENDOR_INFO_T */

e1000_vendor_info_t e1000_vendor_info_array[] = {
	{0x8086, 0x1000, "Intel(R) PRO/1000 Gigabit Server Adapter"},
	{0x8086, 0x1003, "Intel(R) PRO/1000 Gigabit Adapter (SC - Fiber)"},
	{CATCHALL, CATCHALL, "Intel(R) 8254x Based Gigabit Adapter"},
	{0x0, 0x0, ""}					  /* This has to be the last entry */
};

